import torch

from .segmentation import MCD_loss, FocalLoss
from .regression import CCE_Generic, CE, CE_Logits


def DCCE(prediction: torch.Tensor, target: torch.Tensor, params: dict) -> torch.Tensor:
    """
    Calculates the Dice-Cross-Entropy loss.

    Args:
        prediction (torch.Tensor): The predicted mask.
        target (torch.Tensor): The ground truth mask.
        params (dict): The parameters.

    Returns:
        torch.Tensor: The calculated loss.
    """
    dcce_loss = MCD_loss(prediction, target, params) + CCE_Generic(
        prediction, target, params, CE
    )
    return dcce_loss


def DCCE_Logits(
    prediction: torch.Tensor, target: torch.Tensor, params: dict
) -> torch.Tensor:
    """
    Calculates the Dice-Cross-Entropy loss using logits.

    Args:
        prediction (torch.Tensor): The predicted mask.
        target (torch.Tensor): The ground truth mask.
        params (dict): The parameters.

    Returns:
        torch.Tensor: The calculated loss.
    """
    dcce_loss = MCD_loss(prediction, target, params) + CCE_Generic(
        prediction, target, params, CE_Logits
    )
    return dcce_loss


def DC_Focal(
    prediction: torch.Tensor, target: torch.Tensor, params: dict
) -> torch.Tensor:
    """
    Calculates the Dice-Focal loss.

    Args:
        prediction (torch.Tensor): The predicted mask.
        target (torch.Tensor): The ground truth mask.
        params (dict): The parameters.

    Returns:
        torch.Tensor: The calculated loss.
    """
    return MCD_loss(prediction, target, params) + FocalLoss(prediction, target, params)



import logging
import yaml
from pathlib import Path
from importlib import resources
import tempfile
from GANDLF.utils import get_unique_timestamp
import sys


def _create_tmp_log_file():
    tmp_dir = Path(tempfile.gettempdir())
    log_dir = Path.joinpath(tmp_dir, ".gandlf")
    log_dir.mkdir(parents=True, exist_ok=True)
    log_file = Path.joinpath(log_dir, get_unique_timestamp() + ".log")
    return log_file


def _create_log_file(log_file):
    log_file = Path(log_file)
    log_file.write_text("Starting GaNDLF logging session \n")


def _configure_logging_with_logfile(log_file, config_path):
    with resources.open_text("GANDLF", config_path) as file:
        config_dict = yaml.safe_load(file)
        config_dict["handlers"]["rotatingFileHandler"]["filename"] = str(log_file)
        logging.config.dictConfig(config_dict)


def gandlf_excepthook(exctype, value, tb):
    if issubclass(exctype, Exception):
        logging.exception("Uncaught exception", exc_info=(exctype, value, tb))
    else:
        sys.__excepthook__(exctype, value, tb)


def logger_setup(log_file=None, config_path="logging_config.yaml") -> None:
    """
    It sets up the logger. Reads from logging_config.

    Args:
        log_file (str): dir path for saving the logs, defaults to `None`, at which time logs are flushed to console.
        config_path (str): file path for the configuration

    """

    logging.captureWarnings(True)
    log_tmp_file = log_file
    if log_file is None:  # create tmp file
        log_tmp_file = _create_tmp_log_file()
    _create_log_file(log_tmp_file)
    _configure_logging_with_logfile(log_tmp_file, config_path)
    sys.excepthook = gandlf_excepthook
    logging.info(f"The logs are saved in {log_tmp_file}")


class InfoOnlyFilter(logging.Filter):
    """
    Display only INFO messages.
    """

    def filter(self, record):
        """
        Determines if the specified record is to be logged.
        Args:
            record (logging.LogRecord): The log record to be evaluated.
        Returns:
            bool: True if the log record should be processed, False otherwise.
        """
        return record.levelno == logging.INFO



#!usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import ast
import logging

import click
from deprecated import deprecated
from GANDLF.cli import preprocess_and_save, copyrightMessage
from GANDLF.entrypoints import append_copyright_to_help
from GANDLF.utils import logger_setup


def _preprocess(
    config: str,
    input_data: str,
    output_dir: str,
    label_pad: str,
    apply_augs: bool,
    crop_zero: bool,
):
    print(f"{config=}")
    print(f"{input_data=}")
    print(f"{output_dir=}")
    print(f"{label_pad=}")
    print(f"{apply_augs=}")
    print(f"{crop_zero=}")
    preprocess_and_save(
        data_csv=input_data,
        config_file=config,
        output_dir=output_dir,
        label_pad_mode=label_pad,
        applyaugs=apply_augs,
        apply_zero_crop=crop_zero,
    )

    # TODO: in `old_way` default logging level is warning, thus those 'finished' are not printed anymore
    logging.info("Finished.")


@click.command()
@click.option(
    "--config",
    "-c",
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
    required=True,
    help="The configuration file (contains all the information related to the training/inference session),"
    " this is read from 'output' during inference",
)
@click.option(
    "--input-data",
    "-i",  # TODO: mention pickled df also fits
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
    required=True,
    help="Data csv file that is used for training/inference",
)
@click.option(
    "--output-dir",
    "-o",
    type=click.Path(file_okay=False, dir_okay=True),
    required=True,
    help="Output directory to save intermediate files and model weights",
)
@click.option(
    "--label-pad",
    "-l",
    type=str,
    default="constant",
    help="Specifies the padding strategy for the label when 'patch_sampler' is 'label'. "
    "Defaults to 'constant' [full list: https://numpy.org/doc/stable/reference/generated/numpy.pad.html]",
)
@click.option(
    "--apply-augs",
    "-a",
    is_flag=True,
    help="If passed, applies data augmentations during output creation",
)
@click.option(
    "--crop-zero",
    "-z",
    is_flag=True,
    help="If passed, applies zero cropping during output creation.",
)
@click.option(
    "--log-file",
    type=click.Path(),
    default=None,
    help="Output file which will contain the logs.",
)
@append_copyright_to_help
def new_way(
    config: str,
    input_data: str,
    output_dir: str,
    label_pad: str,
    apply_augs: bool,
    crop_zero: bool,
    log_file: str,
):
    """Generate training/inference data which are preprocessed to reduce resource footprint during computation."""

    logger_setup(log_file)
    _preprocess(
        config=config,
        input_data=input_data,
        output_dir=output_dir,
        label_pad=label_pad,
        apply_augs=apply_augs,
        crop_zero=crop_zero,
    )


@deprecated(
    "This is a deprecated way of running GanDLF. Please, use `gandlf preprocess` cli command "
    + "instead of `gandlf_preprocess`. Note that in new CLI tool some params were renamed to snake-case:\n"
    + "  --inputdata to --input-data\n"
    + "  --labelPad to --label-pad\n"
    + "  --applyaugs to --apply-augs; it is flag now, i.e. no value accepted\n"
    + "  --cropzero to --crop-zero;  it is flag now, i.e. no value accepted\n"
    + "`gandlf_preprocess` script would be deprecated soon."
)
def old_way():
    logger_setup()
    parser = argparse.ArgumentParser(
        prog="GANDLF_Preprocess",
        formatter_class=argparse.RawTextHelpFormatter,
        description="Generate training/inference data which are preprocessed to reduce resource footprint during computation.\n\n"
        + copyrightMessage,
    )
    parser.add_argument(
        "-c",
        "--config",
        metavar="",
        type=str,
        help="The configuration file (contains all the information related to the training/inference session), this is read from 'output' during inference",
        required=True,
    )
    parser.add_argument(
        "-i",
        "--inputdata",
        metavar="",
        type=str,
        help="Data csv file that is used for training/inference",
        required=True,
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="",
        type=str,
        help="Output directory to save intermediate files and model weights",
        required=True,
    )
    parser.add_argument(
        "-l",
        "--labelPad",
        metavar="",
        type=str,
        default="constant",
        help="This specifies the padding strategy for the label when 'patch_sampler' is 'label'. Defaults to 'constant' [full list: https://numpy.org/doc/stable/reference/generated/numpy.pad.html]",
        required=False,
    )
    parser.add_argument(
        "-a",
        "--applyaugs",
        metavar="",
        type=ast.literal_eval,
        default=False,
        help="This specifies the whether to apply data augmentation during output creation. Defaults to False",
        required=False,
    )
    parser.add_argument(
        "-z",
        "--cropzero",
        metavar="",
        type=ast.literal_eval,
        default=False,
        help="This specifies the whether to apply zero cropping during output creation. Defaults to False",
        required=False,
    )

    args = parser.parse_args()
    _preprocess(
        config=args.config,
        input_data=args.inputdata,
        output_dir=args.output,
        label_pad=args.labelPad,
        apply_augs=args.applyaugs,
        crop_zero=args.cropzero,
    )



# -*- coding: utf-8 -*-
"""
Modified from https://github.com/pytorch/vision.git
"""

import math
import torch.nn as nn
import torch.nn.functional as F

from .modelBase import ModelBase


class VGG(ModelBase):
    """
    VGG model
    """

    def __init__(self, parameters: dict, configuration):
        """
        Initializer function for the VGG model

        Args:
            parameters (dict) - overall parameters dictionary; parameters specific for DenseNet:
            configuration (dict): A dictionary of configuration parameters for the model.

        """
        super(VGG, self).__init__(parameters)

        # amp is not supported for vgg
        parameters["model"]["amp"] = False

        # Setup the feature extractor
        self.features = self.make_layers(configuration, self.n_channels)

        # Setup the classifier
        # Dev Note: number of input features for linear layer should be changed later,
        # but works for all vgg right now
        self.classifier = nn.Sequential(
            nn.Dropout(),
            self.GlobalAvgPool(),
            nn.ReLU(True),
            nn.Dropout(),
            nn.Linear(512, self.n_classes),
        )

        # Initialize weights, if convolutional use He initialization, if linear use Xavier initialization
        for m in self.modules():
            if isinstance(m, self.Conv):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2.0 / n))
                m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight)
            else:
                pass

    def forward(self, x):
        """
        Forward pass function of the VGG model.

        Args:
            x (tensor): Input tensor of shape (batch_size, n_channels, height, width).

        Returns:
            tensor: Output tensor of shape (batch_size, n_classes).
        """
        out = self.features(x)
        out = self.classifier(out)
        if not self.final_convolution_layer is None:
            if self.final_convolution_layer == F.softmax:
                # Apply softmax activation to the output tensor if specified in the configuration
                out = self.final_convolution_layer(out, dim=1)
            else:
                # Apply whatever final layer specified in the configuration to the output tensor
                out = self.final_convolution_layer(out)
        return out

    def make_layers(self, layer_config, input_channels):
        """
        Function to create convolutional layers for the VGG model based on the given layer configuration
        for VGG based models including VGG11, VGG13, VGG16, VGG19.

        Args:
            layer_config (list): A list containing the configuration of convolutional layers and max pooling layers
            in_channels (int): The number of input channels

        Returns:
            nn.Sequential: A sequential module containing the convolutional layers
        """
        layers = []
        for layer in layer_config:
            if layer == "M":
                # If you found M, then add a maxpool layer
                layers += [self.MaxPool(kernel_size=2, stride=2)]
            else:
                # Otherwise, add a convolutional layer with the number of channels
                conv = self.Conv(input_channels, layer, kernel_size=3, padding=1)
                if self.norm_type in ["batch", "instance"]:
                    layers += [conv, self.Norm(layer), nn.ReLU(inplace=True)]
                else:
                    layers += [conv, nn.ReLU(inplace=True)]
                input_channels = layer
        return nn.Sequential(*layers)


# Layer configuration for VGG models, as per the paper and M represents maxpool
# and the integers represent the number of channels in the convolutional layers
cfg = {
    "A": [64, "M", 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M"],
    "B": [64, 64, "M", 128, 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M"],
    "D": [
        64,
        64,
        "M",
        128,
        128,
        "M",
        256,
        256,
        256,
        "M",
        512,
        512,
        512,
        "M",
        512,
        512,
        512,
        "M",
    ],
    "E": [
        64,
        64,
        "M",
        128,
        128,
        "M",
        256,
        256,
        256,
        256,
        "M",
        512,
        512,
        512,
        512,
        "M",
        512,
        512,
        512,
        512,
        "M",
    ],
}


class vgg11(VGG):
    """
    A class representing the VGG11 model, which is a variant of the VGG model architecture.

    Inherits from the VGG class and specifies the configuration for the VGG11 architecture.
    """

    def __init__(self, parameters):
        """
        Initializes the VGG11 model with the given parameters.

        Args:
            parameters (dict): A dictionary containing the parameters for the VGG11 model.
        """
        super(vgg11, self).__init__(parameters=parameters, configuration=cfg["A"])


class vgg13(VGG):
    """
    A class representing the VGG13 model, which is a variant of the VGG model architecture.

    Inherits from the VGG class and specifies the configuration for the VGG13 architecture.
    """

    def __init__(self, parameters):
        """
        Initializes the VGG13 model with the given parameters.

        Args:
            parameters (dict): A dictionary containing the parameters for the VGG13 model.
        """
        super(vgg13, self).__init__(parameters=parameters, configuration=cfg["B"])


class vgg16(VGG):
    """
    A class representing the VGG16 model, which is a variant of the VGG model architecture.

    Inherits from the VGG class and specifies the configuration for the VGG16 architecture.
    """

    def __init__(self, parameters):
        """
        Initializes the VGG16 model with the given parameters.

        Args:
            parameters (dict): A dictionary containing the parameters for the VGG16 model.
        """
        super(vgg16, self).__init__(parameters=parameters, configuration=cfg["D"])


class vgg19(VGG):
    """
    A class representing the VGG19 model, which is a variant of the VGG model architecture.

    Inherits from the VGG class and specifies the configuration for the VGG19 architecture.
    """

    def __init__(self, parameters):
        """
        Initializes the VGG19 model with the given parameters.

        Args:
            parameters (dict): A dictionary containing the parameters for the VGG19 model.
        """
        super(vgg19, self).__init__(parameters=parameters, configuration=cfg["E"])
